require "defines"
require "utils"

-- Energy used per distance per item teleported per tick.
energyPerItem = 10
-- The length of time that the energy is used, this smooths out the usage a bit.
energyUsageTime = 100

remote.addinterface("TeleBelts",
{
	printTelepairs = function()
		for name, pair in pairs(glob.teles.telepairs) do
			game.player.print(name)
		end
	end,
	clean = function()
		glob.teles = {	guiTimer = 0,
						guiOpen = false,
						guiPairName = nil,
						guiEntity = nil,
						telepairs = {},
						version = utils.currentVersion
					 }
	end
})

game.oninit(function() OnInit() end)
game.onload(function() OnLoad() end)
game.onevent(defines.events.ontick, function(event) OnTick(event) end)
game.onevent(defines.events.onbuiltentity, function(event) OnBuiltEntity(event) end)
game.onevent(defines.events.onentitydied, function(event) OnEntityDied(event) end)
game.onevent(defines.events.onpreplayermineditem, function(event) OnPrePlayerMinedItem(event) end)
game.onevent(defines.events.onguiclick, function(event) OnGUIClick(event) end)

-- Lookup table for the position offsets of item drops on transport belts.
beltOffsetsByDirection = {[0] = {left = {x = -0.24, y =  0   }, right = {x =  0.24, y =  0   }},
						  [2] = {left = {x =  0   , y = -0.24}, right = {x =  0   , y =  0.24}},
						  [4] = {left = {x =  0.24, y =  0   }, right = {x = -0.24, y =  0   }},
						  [6] = {left = {x =  0   , y =  0.24}, right = {x =  0   , y = -0.24}}
						 }

function OnInit()
	--game.player.print("TeleBelts: OnInit")
	if glob.teles == nil then
		glob.teles = {	guiTimer = 0,
						guiOpen = false,
						guiPairName = nil,
						guiEntity = nil,
						telepairs = {},
						version = utils.currentVersion
					 }
	end
	if not glob.teles.version then
		game.player.print("Migrating TeleBelts from nonversioned to version " .. utils.currentVersion)
		utils.migrations.versionless()
	end
	if glob.teles.version ~= utils.currentVersion then
		game.player.print("Migrating TeleBelts from version " .. glob.teles.version .. " to version " .. utils.currentVersion .. ".")
		utils.migrations[glob.teles.version]()
		game.player.print("Migration successful.")
	end
end

function OnLoad()
	--game.player.print("TeleBelts: OnLoad")
	-- Call OnInit to ensure glob.teles exists.
	OnInit()
end
						  
function OnTick(_Event)
	--game.player.print("TeleBelts: OnTick")
	-- Selection has to be done though a hacky hover, since custom entity OnClick is not possible yet.
	for index, player in ipairs(game.players) do
		if game.player.selected ~= nil then
			if IsTele(game.player.selected) then
				glob.teles.guiTimer = glob.teles.guiTimer + 1
			end
			
			if glob.teles.guiTimer >= 100 then
				OpenGUI(player.selected,index)
			end
		else
			glob.teles.guiTimer = 0
		end
	end
	
	for name, telepair in pairs(glob.teles.telepairs) do
		if telepair.receiver and telepair.receiver.valid and telepair.sender and telepair.sender.valid then
		
			-- Use the energy to be used from previous ticks.
			telepair.energyUsage[game.tick % energyUsageTime + 1] = 0
			for i, energy in pairs(telepair.energyUsage) do
				telepair.sender.energy = telepair.sender.energy - energy
			end
			
			
			-- If there's enough energy left.
			if telepair.sender.energy > energyPerItem * telepair.distSq then
				--game.player.print("Handling pair " .. name)
				local valid = false
				-- Check if there's still items in the stored stacks and if they're still valid.
				for i, stack in pairs(telepair.stacks) do
					if stack.valid then
						valid = true
						break
					end
				end
				
				-- If they're not, search for new stacks inside this tile.
				telepair.stacks = game.findentitiesfiltered{name = "item-on-ground",
									  				 		area = {{telepair.sender.position.x - 0.5, telepair.sender.position.y - 0.5},
															 		{telepair.sender.position.x + 0.5, telepair.sender.position.y + 0.5}}
														   }
				--game.player.print("Found " .. #telepair.stacks .. " entities")
			
				--game.player.print("Looping over " .. #telepair.stacks .. " entities")
				local teleportedAny = false
				local maxI = 0
				for i, stack in pairs(telepair.stacks) do
					maxI = i
					if stack.valid then
						-- Relative position.
						local r = {x = (stack.position.x - telepair.sender.position.x), y = (stack.position.y - telepair.sender.position.y)}
						
						-- Determine whether the item that entered is on the left or right side of the belt.
						local wasOnLeft = math.abs(r.x) > math.abs(r.y)
						if r.x * r.y < 0 then wasOnLeft = not wasOnLeft end
						
						-- Get the required offset at the target location.
						local off = beltOffsetsByDirection[telepair.receiver.direction].right
						if wasOnLeft then
							off = beltOffsetsByDirection[telepair.receiver.direction].left
						end
						
						-- Calculate position to drop the item.
						local newPos = {telepair.receiver.position.x + off.x, telepair.receiver.position.y + off.y}
						
						-- Check if the target location is empty.
						local possiblePos = game.findnoncollidingposition("item-on-ground", newPos, 0.10, 0.01)
						if possiblePos then
							telepair.energyUsage[game.tick % energyUsageTime + 1] = telepair.energyUsage[game.tick % energyUsageTime + 1] + energyPerItem * telepair.distSq
							--telepair.sender.energy = telepair.sender.energy - energyPerItem * telepair.distSq
	
							game.createentity{position=possiblePos, name="item-on-ground", stack={name=stack.stack.name, count=stack.stack.count}}
							stack.destroy()
							--stack.teleport(possiblePos)
	
							teleportedAny = true
						end
					end
				end
				
				-- If none were teleported, it's possible one lane is full, so force rescan.
				if not teleportedAny then
					telepair.stacks = {}
				end
			end
		end
	end
end

function OnBuiltEntity(_Event)
	--game.player.print("TeleBelts: OnBuiltEntity")
	if IsTele(_Event.createdentity) then
		if glob.teles.guiOpen then
			CloseGUI(game.players[_Event.playerindex])
		end
		OpenGUI(_Event.createdentity,_Event.playerindex)
	end
end

function OnEntityDied(_Event)
	--game.player.print("TeleBelts: OnEntityDied")
	if not IsTele(_Event.entity) then return end
	
	RemoveReceiverOrSender(_Event.entity,1)--1 is dirty hack 
end

function OnPrePlayerMinedItem(_Event)
	--game.player.print("TeleBelts: OnPrePlayerMinedItem")
	if not IsTele(_Event.entity) then return end
	
	RemoveReceiverOrSender(_Event.entity,_Event.playerindex)
end


function OnGUIClick(_Event)
	--game.player.print("TeleBelts: OnGUIClick")
	if _Event.element.name == "tb_okbutton" then
		local oldName = glob.teles.guiPairName
		local newName = game.players[_Event.playerindex].gui.center.tb_frame.tb_textfield.text
		if string.len(newName) > 0 then
			if not glob.teles.telepairs[oldName] then
				game.players[_Event.playerindex].print("TeleBelts: Clicked OK in gui with invalid glob.teles.guiPairName (" .. oldName .. ").")
			end
			
			ChangeReceiverOrSenderTelepair(glob.teles.guiEntity, newName, oldName,_Event.playerindex)
		else
			game.players[_Event.playerindex].print("TeleBelts: Cannot reroute, target name must be at least 1 character long.")
		end
		CloseGUI(game.players[_Event.playerindex])
	end
end

function OpenGUI(_Entity,playerindex)
	--game.player.print("TeleBelts: OpenGUI")
	if glob.teles.guiOpen then
		CloseGUI(game.players[playerindex])
	end
	if not IsTele(_Entity) then return nil end
	
	local isSender = _Entity.name == "telesender"
	
	pairName = FindTelepairName(_Entity)
	if pairName == nil then
		pairName = "def_" .. game.tick
		
		--game.player.print("Created new pair: " .. pairName)
		ChangeReceiverOrSenderTelepair(_Entity, pairName, nil,playerindex)
	end
	
	glob.teles.guiPairName = pairName
	glob.teles.guiEntity = _Entity
	local frame = game.players[playerindex].gui.center.add({type="frame", name="tb_frame", caption="Rename this one", direction="vertical"})
		local textField = frame.add({type="textfield", name="tb_textfield", caption="", style="textfield_style"})
			textField.style.maximalwidth=300
			textField.style.minimalwidth=300
			textField.caption = pairName
			textField.text = pairName
		frame.add{type = "button", name = "tb_okbutton", caption = "Confirm", style = "button_style"}
	glob.teles.guiOpen = true
end

function CloseGUI(player)
	--game.player.print("TeleBelts: CloseGUI")
	if not glob.teles.guiOpen then
		player.print("TeleBelts: Tried to close GUI while it's not open.")
		return
	end
	if player.gui.center.tb_frame ~= nil then
		player.gui.center.tb_frame.destroy()
	end
	glob.teles.guiPairName = nil
	glob.teles.guiEntity = nil
end

function RemoveReceiverOrSender(_Entity,playerindex)
	if not IsTele(_Entity) then
		game.players[playerindex].print("TeleBelts: RemoveReceiverOrSender received entity " .. _Entity.name .. ".")
		return
	end
	local pairName = FindTelepairName(_Entity)
	
	tp, nottype = GetTeleType(_Entity)
	
	if not glob.teles.telepairs[pairName] then
		game.players[playerindex].print("TeleBelts: RemoveReceiverOrSender received invalid telepair name " .. pairName .. ".")
		return
	end
	
	glob.teles.telepairs[pairName][tp] = nil
	CleanPair(pairName)
end
	
	
	


function ChangeReceiverOrSenderTelepair(_Entity, _NewPairName, _OldPairName,playerindex)
	if not IsTele(_Entity) then
		game.players[playerindex].print("TeleBelts: ChangeTelepairReceiverOrSender received invalid entity " .. _Entity.name .. ".")
		return
	end
	
	if _OldPairName == _NewPairName then return end
	
	
	if not glob.teles.telepairs[_NewPairName] then
		glob.teles.telepairs[_NewPairName] = {	sender = nil,
											  	receiver = nil,
											  	distSq = 0,
												stacks = {},
												energyUsage = {}
											 }
	end
	
	local newPair = glob.teles.telepairs[_NewPairName]
	
	tp, nottype = GetTeleType(_Entity)
	
	if newPair[tp] then
		game.players[playerindex].print("TeleBelts: Cannot reroute " .. tp .. ", target name (" .. _NewPairName .. ") already taken.")
		return
	end
	
	-- If the old pair exists, remove it from that pair and clean the pair.
	if _OldPairName and glob.teles.telepairs[_OldPairName] then
		glob.teles.telepairs[_OldPairName][tp] = nil
		CleanPair(_OldPairName)
	end
	
	newPair[tp] = _Entity
	
	-- Calculate squared distance between sender and receiver.
	if newPair[nottype] then
		--game.player.print("type = " .. tp .. ", nottype = " .. nottype)
		
		local diff = {x = newPair[tp].position.x - newPair[nottype].position.x, y = newPair[tp].position.y - newPair[nottype].position.y}
		newPair.distSq = diff.x * diff.x + diff.y * diff.y
	end
end

function FindTelepairName(_Entity)
	--game.player.print("TeleBelts: FindTelepairName")
	if not _Entity.valid then return nil end
	
	if _Entity.name == "telesender" then
		for name, telepair in pairs(glob.teles.telepairs) do
			if telepair.sender and telepair.sender.valid then
				if telepair.sender.position.x == _Entity.position.x and telepair.sender.position.y == _Entity.position.y then
					return name
				end
			end
		end
	elseif _Entity.name == "telereceiver" then
		for name, telepair in pairs(glob.teles.telepairs) do
			if telepair.receiver and telepair.receiver.valid then
				if telepair.receiver.position.x == _Entity.position.x and telepair.receiver.position.y == _Entity.position.y then
					return name
				end
			end
		end
	end
	--game.player.print("TeleBelts: FindTelepairName: Not found.")
	return nil
end

function CleanPair(_PairName,playerindex)
	if not glob.teles.telepairs[_PairName] then
		game.players[playerindex].print("TeleBelts: CleanPair received nonexistent pair (" .. _PairName .. ")")
		return
	end
	
	if not glob.teles.telepairs[_PairName].sender or not glob.teles.telepairs[_PairName].sender.valid then
		if not glob.teles.telepairs[_PairName].receiver or not glob.teles.telepairs[_PairName].receiver.valid then
			-- No valid sender and no valid receiver. Delete the pair.
			glob.teles.telepairs[_PairName] = nil
		end
	end
end

function GetTeleType(_Entity)
	local tp = "sender"
	local nottype = "receiver"
	if _Entity.name == "telereceiver" then
		tp = "receiver"
		nottype = "sender"
	end
	
	return tp, nottype
end

function IsTele(_Entity)
	if not _Entity or not _Entity.valid then
		return false
	end
	if _Entity.name == "telesender" or _Entity.name == "telereceiver" then
		return true
	end
	return false
end
